from typing import List, Dict, Any
from flitsr.spectrum import Spectrum, Outcome


class SpectrumBuilder:
    def __init__(self):
        self._elements: List[Spectrum.Element] = []
        self._tests: List[Spectrum.Test] = []
        self._groups: List[Spectrum.Group] = [Spectrum.Group()]
        self._executions: Dict[Spectrum.Test, Dict[Spectrum.Element, bool]] = {}

    def get_tests(self):
        return self._tests

    def addTest(self, name: str, index: int, outcome: Outcome):
        t = Spectrum.Test(name, index, outcome)
        self._tests.append(t)
        self._executions[t] = {}
        return t

    def addElement(self, details: List[str],
                   faults: List[Any]) -> Spectrum.Element:
        e = Spectrum.Element(details, len(self._elements), faults)
        self._elements.append(e)
        self._groups[0].append(e)
        return e

    def addExecution(self, test: Spectrum.Test, elem: Spectrum.Element,
                     executed: bool):
        if (test not in self._executions):
            self._executions[test] = {}
        self._executions[test][elem] = executed
        # if (executed):
        #     if (test.outcome is Outcome.PASSED):
        #         self.p[elem] += 1
        #     else:
        #         self.f[elem] += 1

    def split_groups_on_test(self, test: Spectrum.Test):
        """
        Given one test pertaining to a row in the spectrum, split the groups
        according to the coverage. The split will potentially split each group
        in half, based on those elements that are executed in the test, and
        those that are not.
        """
        row = self._executions[test]
        new_groups: List[Spectrum.Group] = []
        for group in self._groups:
            elems = group.get_elements()
            collect: List[List[Spectrum.Element]] = [[], []]
            for elem in elems:
                collect[row.get(elem, False)].append(elem)
            for c in collect:
                if (c != []):
                    new_groups.append(Spectrum.Group(c))
        self._groups = new_groups

    def get_spectrum(self):
        spectrum = Spectrum(self._elements, self._groups, self._tests,
                            self._executions)
        return spectrum

    # def remove_unnecessary(self):
    #     """
    #     Remove the unnecessary elements that are within the groups from the
    #     spectrum.
    #     """
    #     remove = []
    #     for group in self._groups:
    #         remove.extend(group[1:])
    #         self.group_dict[group[0]] = group
    #     # remove.sort(reverse=True)
    #     # self.locs -= len(remove)
    #     for elem in remove:
    #         self.remove_element(elem)



import sys
from flitsr.spectrum import Spectrum
from typing import Dict, List, Set, Tuple


def split(faults: Dict[float, List[Spectrum.Element]],
          spectrum: Spectrum) -> Tuple[Dict[float, List[Spectrum.Element]],
                                       Set[Spectrum.Element]]:
    """
    Splits fault groups that are a combination of two or more sub-faults in
    mutually exclusive parts of the system into separate faults. Also finds
    faults that are not exposed and returns them separately.
    """
    if (faults == {}):
        return {}, set()
    ftemp = [([elem], f[0], False) for f in faults.items() for elem in f[1]]
    for test in spectrum.failing():
        merge: Dict[float, List[Spectrum.Element]] = {}
        remain = []
        for equiv in ftemp:
            if (any(spectrum[test][e] for e in equiv[0])):
                merge.setdefault(equiv[1], []).extend(equiv[0])
            else:
                remain.append(equiv)
        if (len(merge) != 0):
            for (f_num, f_locs) in merge.items():
                remain.append((f_locs, f_num, True))
        ftemp = remain
    fmap: Dict[float, List[List[Spectrum.Element]]] = {}
    unexposed: Set[Spectrum.Element] = set()
    for equiv in ftemp:
        if (not equiv[2]):
            unexposed.update(equiv[0])
            continue
        fmap.setdefault(equiv[1], []).append(equiv[0])
    new_faults: Dict[float, List[Spectrum.Element]] = {}
    for item2 in fmap.items():
        if (len(item2[1]) == 1):
            new_faults[item2[0]] = item2[1][0]
        else:
            for i in range(len(item2[1])):
                new_faults[float("{}.{}".format(item2[0], i+1))] = item2[1][i]
    return new_faults, unexposed


if __name__ == "__main__":
    d = sys.argv[1]
    i = 2
    gzoltar = False
    num_only = False
    while (True):
        if (len(sys.argv) > i):
            if (sys.argv[i] == "gzoltar"):
                gzoltar = True
            elif (sys.argv[i] == "num"):
                num_only = True
            else:
                print("Unknown option:", sys.argv[i])
                quit()
            i += 1
        else:
            break
    if (gzoltar):
        from flitsr.input import read_spectrum
    else:
        from flitsr.tcm_input import read_spectrum
    spectrum = read_spectrum(d, False)
    faults = spectrum.get_faults()
    print("faults:", faults)
    # print(groups)
    # print_spectrum(spectrum)
    faults, unexposed = split(faults, spectrum)
    print("split faults:", faults)
    print("unexposed:", unexposed)



from flitsr.spectrum import Spectrum
from flitsr.ranking import Ranking
from flitsr.advanced.artemis_impl import artemis
from flitsr.advanced.ranker import Ranker


class Artemis(Ranker):
    """
    Run the ARTEMIS technique on the spectrum to produce the ranked lists.
    This option may be combined with FLITSR and parallel to produce a
    hybrid technique.
    """
    def __init__(self, numUniverse: int = 17, maxUniverse: int = 20,
                 p: float = 1e-5):
        self.numUniverse = numUniverse
        self.maxUniverse = maxUniverse
        self.p = p

    def rank(self, spectrum: Spectrum, metric: str) -> Ranking:
        # Set default metric
        if (metric == 'artemis'):
            metric = 'ochiai'
        elements = spectrum.groups()
        matrix, errVector = spectrum.to_matrix()
        rankingList = artemis.explorer(matrix, errVector, spectrum.locs(),
                                       metric, self.numUniverse,
                                       self.maxUniverse, self.p)
        sortedRankingList = artemis.merge(rankingList, spectrum.locs())
        ranking = Ranking()
        for (elem_ind, score, _) in sortedRankingList:
            ranking.append(elements[int(elem_ind)], score, 0)
        return ranking



from flitsr.ranking import Ranking
from flitsr.advanced.ranker import Ranker
from flitsr.spectrum import Spectrum
from flitsr.suspicious import Suspicious


class SBFL(Ranker):
    """
    Disables the FLITSR algorithm so that only the base metric
    is used to produce the ranking. This is equivalent to using the
    base metric as-is, but allows the user to run these metrics
    within the FLITSR framework
    """
    _print_name = 'base'

    _tiebrk_opts = ['exec', 'rndm', 'orig']

    def __init__(self, tiebrk: str = 'orig'):
        self.tiebrk = self._tiebrk_opts.index(tiebrk) + 1
        pass

    def rank(self, spectrum: Spectrum, base_metric: str) -> Ranking:
        return Suspicious.apply_formula(spectrum, base_metric,
                                        tiebrk=self.tiebrk)



from flitsr.tie import Ties


def first(ties: Ties, c: bool):
    if (len(ties.faults) == 0):
        return 0
    return method(ties, 1, collapse=c)


def average(ties: Ties, c: bool):
    if (len(ties.faults) == 0):
        return 0
    return method(ties, len(ties.faults), avg=True, collapse=c)


def median(ties: Ties, c: bool):
    if (len(ties.faults) == 0):
        return 0
    if (len(ties.faults) % 2 == 1):
        return method(ties, int((len(ties.faults)+1)/2), False, c)
    else:
        m1 = method(ties, int(len(ties.faults)/2), False, c)
        m2 = method(ties, int(len(ties.faults)/2)+1, False, c)
        return (m1+m2)/2


def last(ties: Ties, c: bool):
    if (len(ties.faults) == 0):
        return 0
    return method(ties, len(ties.faults), collapse=c)


def nth(ties: Ties, n: int, c: bool):
    if (len(ties.faults) == 0):
        return 0
    return method(ties, min(len(ties.faults), n), collapse=c)

# <---------------- Helper functions --------------->


def method(ties: Ties, target, avg=False, collapse=False,
           worst_effort=False) -> float:
    found = False
    actual = 0
    effort = 0
    efforts = []
    tie_iter = iter(ties)
    while (not found):
        tie = next(tie_iter)
        # print(tie)
        actual += tie.num_faults()
        found = (actual >= target)
        if (avg):
            for j in range(1, tie.num_faults()+1):
                efforts.append(effort + tie.expected_value(j, True, collapse))
        if (not found):
            effort += tie.len(collapse) - tie.num_fault_locs(collapse)
        else:
            k = target + tie.num_faults() - actual
            effort += tie.expected_value(k, True, collapse)
    if (avg):
        return sum(efforts)/target
    else:
        return effort



# PYTHON_ARGCOMPLETE_OK
import re
from flitsr.percent_at_n import combine
import os
import sys
from os import path as osp
from io import TextIOWrapper
from scipy.stats import wilcoxon
from argparse import ArgumentParser, Action, FileType, ArgumentTypeError
import argcomplete
from flitsr.file import File
from flitsr.suspicious import Suspicious
from typing import Set, Dict, List, Tuple, Collection, Optional

PERC_N = "percentage at n"


class Avg:
    """Holds a partially constructed average"""
    def __init__(self, size=None, percn=False):
        self.all = []
        self.adds = 0
        self.percn = percn
        self.rel = False
        if (size is not None):
            self.rel = True
            self.size = size

    def add(self, val):
        self.all.append(val)
        self.adds += 1

    def eval(self):
        if (self.percn):
            return self.all
        elif (self.rel):
            return sum(s*100/self.size for s in self.all)/self.adds
        else:
            return sum(self.all)/self.adds

    def significance(self, avg: 'Avg'):
        # if number of observations differ, no significance
        if (len(self.all) != len(avg.all)):
            return ('equal', 1.0)
        diff = 1.0 if (self.all == avg.all) else \
            wilcoxon(self.all, avg.all, "pratt", method="approx").pvalue
        if (diff >= 0.05):
            return ('equal', diff)
        else:
            greater = wilcoxon(self.all, avg.all, "pratt", method="approx",
                               alternative="greater").pvalue
            less = wilcoxon(self.all, avg.all, "pratt", method="approx",
                            alternative="less").pvalue
        if (greater < 0.05):
            return ('greater', greater)
        else:
            return ('less', less)


def readBlock(file: File) -> List[str]:
    block = []
    line = file.readline()
    if (": " not in line):
        line = file.readline()
    while (line != "" and not re.match("^-+$", line)):
        block.append(line)
        line = file.readline()
    return block


def find_dirs(dirs: List[str], path, depth=1, max=None,
              incl=Dict[str, List[str]], excl=Dict[str, List[str]]):
    """
    Recursively traverse a directory structure at the given `path` and find
    directories containing results files, adding them to `dirs`. An optional
    maximum depth may be given, along with names to include and exclude at
    various depth levels ("*" indicates any level).
    """
    # Get the names to include and exclude at this depth level
    incl_lvl = incl.get(str(depth), []) + incl.get("*", [])
    excl_lvl = excl.get(str(depth), []) + excl.get("*", [])
    for dir in os.scandir(path):
        if (dir.is_dir() and (dir.name not in excl_lvl) and
            (len(incl_lvl) == 0 or dir.name in incl_lvl)):
            new_path = dir.path
            if ((max and depth >= max) or
                (not max and any(f.endswith('.results') for f in
                                 os.listdir(new_path)))):
                dirs.append(new_path+"/")
            else:
                find_dirs(dirs, new_path, depth=depth+1, max=max, incl=incl,
                          excl=excl)


def ci_eq(str1: str, str2: str):
    """ Case insensitive equality check for strings """
    return str1.casefold() == str2.casefold()


def ci_in(str1: str, col: Collection[str]):
    """ Case insensitive check for strings in a list """
    return str1.casefold() in map(str.casefold, col)


def merge(recurse: bool, max: int, incl: List[Tuple[str, str]],
          excl: List[Tuple[str, str]], rel: bool,
          output_file: TextIOWrapper, perc_file: TextIOWrapper,
          tex_file: TextIOWrapper, dec=2, group='metric', incl_calcs=None,
          percs=None, incl_metrics=None, flitsrs=None, sign=None,
          sign_less=[], base_type=None, thrs=[], keep_order=False):
    # Set up the include and exclude dir names dicts
    incl_dict: Dict[str, List[str]] = {}
    excl_dict: Dict[str, List[str]] = {}
    for d, n in incl:
        incl_dict.setdefault(d, []).append(n)
    for d, n in excl:
        excl_dict.setdefault(d, []).append(n)
    # set up other variables
    if (base_type is None):
        base_type = 'base'
    temp_metrics: Set[str] = set()
    modes: Set[str] = set()
    temp_calcs: Set[str] = set()
    #           dir       mode      metric
    files: Dict[str, Dict[str, Dict[str, File]]] = {}
    #          mode      metric    calc
    avgs: Dict[str, Dict[str, Dict[str, Avg]]] = {}
    if (rel):
        if (not recurse):
            size = int(open("../size").readline())
    # Find the directories and results files
    dirs = [""]
    if (recurse):
        dirs = []
        find_dirs(dirs, ".", max=max, incl=incl_dict, excl=excl_dict)
    # Read in all results
    if (rel):
        sizes = {}
    results_check = re.compile("^(?:([\\w_]*)_)?(\\w+)\\.results$")
    for d in dirs:
        files.setdefault(d, {})
        for file in os.scandir(osp.normpath(d)):
            m = results_check.match(file.name)
            if (m):
                mode = m.group(1) or ""
                modes.add(mode)
                metric = m.group(2)
                temp_metrics.add(metric)
                files[d].setdefault(mode, {})[metric] = File(file)
                avgs.setdefault(mode, {}).setdefault(metric, {})
        if (rel):
            sizes[d] = int(open(osp.join(d, "../size")).readline())

    # Collect all the results
    for d in dirs:
        for mode in modes:
            for metric in temp_metrics:
                try:
                    block = readBlock(files[d][mode][metric])
                except Exception:
                    print('WARNING: Could not find results for '
                          f'{mode.replace("_", " ").title()} '
                          f'{metric.replace("_", " ").title()} in dir {d}',
                          file=sys.stderr)
                    continue
                while (block != []):
                    for line in block:
                        line_s = line.split(": ")
                        calc = line_s[0]
                        temp_calcs.add(calc)
                        if (calc == PERC_N):
                            vals = line_s[1].split(",")
                            avgs[mode][metric].setdefault(calc, Avg(percn=True)).add(
                                (float(vals[0]), [float(x) for x in vals[1:]]))
                        else:
                            s = None
                            if (rel):
                                s = sizes[d] if recurse else size
                            avgs[mode][metric].setdefault(calc, Avg(s)).add(
                                    float(line_s[1]))
                    block = readBlock(files[d][mode][metric])
    # calculate thresholds
    for thr in thrs:
        thr_key = str(thr).lower()
        temp_calcs.add(thr_key)
        for mode in modes:
            for metric in temp_metrics:
                vals = avgs[mode][metric][thr.calc].all
                count = 0.0
                for val in vals:
                    if (thr.comp(val, thr.threshold)):
                        count += 1
                avgs[mode][metric][thr_key] = Avg()
                avgs[mode][metric][thr_key].adds = len(vals)
                avgs[mode][metric][thr_key].all = [count]
    # select only included calcs
    if (incl_calcs is not None):
        # preserves the order of calcs from cmd line
        calcs = [c for c in incl_calcs if ci_in(c, temp_calcs)]
    else:
        calcs = list(temp_calcs)
    # select only included metrics
    if (incl_metrics is not None):
        # preserves the order of metrics from cmd line
        metrics = [m for m in incl_metrics if ci_in(m, temp_metrics)]
    else:
        metrics = list(temp_metrics)
    # sort metrics or keep cmd line order
    if (not keep_order):
        calcs = sorted(calcs)
        metrics = sorted(metrics)

    def print_heading(name, tabs):
        name_disp = name.replace("_", " ").title()
        print('\t'*tabs, name_disp, sep='', file=output_file)
        if (ci_in(PERC_N, calcs) and perc_file is not None and
                perc_file != output_file):
            print('\t'*tabs, name_disp, sep='', file=perc_file)

    def print_tex_heading(mode, metric):
        metric_disp = metric.replace("_", " ").title()
        mode_disp = mode.replace("_", " ").title()
        print('%25s' % (metric_disp + " " + mode_disp), end=" & ",
              file=tex_file)

    def print_results(mode, metric):
        for j, calc in enumerate(calcs):
            if (ci_eq(calc, PERC_N)):
                if (perc_file is not None):
                    comb = combine(avgs[mode][metric][calc].eval())
                    print("\t\t", calc+": ", comb, sep='', file=perc_file)
            else:
                avg = avgs[mode][metric][calc]
                if (not rel and percs is not None and ci_in(calc, percs)):
                    result = round(100*avg.eval(), dec)
                else:
                    result = round(avg.eval(), dec)
                sign_disp = ""
                if (sign is not None):
                    signis: Dict[str, List[Tuple[str, float]]] = {}
                    if (sign == 'type'):
                        for m_alt in modes:
                            if (m_alt == mode or not
                                (ci_in(m_alt, avgs) and
                                 ci_in(metric, avgs[m_alt]))):
                                continue
                            r, p = avg.significance(avgs[m_alt][metric][calc])
                            signis.setdefault(r, []).append((m_alt, p))
                    else:
                        for m_alt in metrics:
                            if (m_alt == metric or not
                                (ci_in(mode, avgs) and
                                 ci_in(m_alt, avgs[mode]))):
                                continue
                            r, p = avg.significance(avgs[mode][m_alt][calc])
                            signis.setdefault(r, []).append((m_alt, p))
                    sign_disp = f" (significantly {signis})"
                print("\t\t", calc+": ", result, sign_disp, sep='',
                      file=output_file)
                if (tex_file):
                    # process TeX significance only for advanced types
                    sign_disp = ''
                    base = base_type.format(re.sub('_[^_]*$', '', mode))
                    if (sign == 'type' and not ci_eq(mode, base)):
                        try:
                            r, p = avg.significance(
                                    avgs[base][metric][calc])
                            if ((calc in sign_less and r == 'less') or
                                (calc not in sign_less and r == 'greater')):
                                sign_disp = '\\tp'
                        except KeyError:
                            pass
                    end = (" & " if j+1 != len(calcs) else " \\\\\n")
                    print('{: <3}'.format(sign_disp),
                          '{: >10.{}f}'.format(result, min(dec, 8)),
                          end=end, file=tex_file)

    # Print out merged results
    if (ci_eq(group, 'metric')):
        zipped = [(mo, me) for me in metrics for mo in sorted(modes)]
    else:
        zipped = [(mo, me) for mo in sorted(modes) for me in metrics]
    cur = None
    # Set up tex file
    if (tex_file):
        print("\\documentclass{standalone}", file=tex_file)
        # print("\\usepackage{longtable}", file=tex_file)
        print("\\begin{document}", file=tex_file)
        # print("\\begin{longtable}", file=tex_file)
        print("\\begin{tabular}{"+'|'.join(['c']*(len(calcs)+1))+"}",
              file=tex_file)
        print("Metric & "+' & '.join([c for c in calcs if not
                                      ci_eq(c, PERC_N)])+"\\\\",
              file=tex_file)
    for (mode, metric) in zipped:
        # check if this mode should be displayed
        if ((ci_eq(mode, 'base') or flitsrs is None or ci_in(metric, flitsrs))
            and (ci_in(mode, avgs) and ci_in(metric, avgs[mode]))):
            if (ci_eq(group, 'metric')):
                if (metric != cur):
                    print_heading(metric, 0)
                    cur = metric
                print_heading(mode, 1)
            else:
                if (mode != cur):
                    print_heading(mode, 0)
                    cur = mode
                print_heading(metric, 1)
            if (tex_file):
                print_tex_heading(mode, metric)
            print_results(mode, metric)
    if (tex_file):
        print("\\end{tabular}", file=tex_file)
        # print("\\end{longtable}", file=tex_file)
        print("\\end{document}", file=tex_file)


def parse_dir_arg(dir_arg: str) -> Tuple[str, str]:
    """
    Parses a given directory argument of the form '[<depth>:]<dir name>',
    where the depth is optional. NOTE: the directory name must not include the
    colon character (':') as this is used to separate the depth.

    Returns a tuple with the depth at which the argument applies, and the
    directory name.
    """
    if (":" in dir_arg):
        depth, _, dir_name = dir_arg.partition(":")
        if (depth != "*" and not depth.isdigit()):
            raise ArgumentTypeError(f'Depth \"{depth}\" not an integer')
    else:
        depth = "*"
        dir_name = dir_arg
    return (depth, dir_name.rstrip('/'))


def above(val, threshold):
    return val >= threshold


def below(val, threshold):
    return val <= threshold


class Threshold:
    def __init__(self, calc, comp, threshold):
        self.calc = calc
        self.comp = comp
        self.threshold = threshold

    def __getitem__(self, i):
        if (i == 0):
            return self.calc
        elif (i == 1):
            return self.comp
        elif (i == 2):
            return self.threshold
        else:
            IndexError("tuple index out of range")

    def __str__(self):
        return f'threshold ({self.calc}, {self.comp.__name__}, {self.threshold})'


def main(argv: Optional[List[str]] = None):
    class RecurseAction(Action):
        def __init__(self, option_strings, dest, nargs=None, **kwargs):
            super().__init__(option_strings, dest, nargs, **kwargs)

        def __call__(self, parser, namespace, values, option_string=None):
            setattr(namespace, self.dest, True)
            if (values is not None):
                setattr(namespace, 'max', values)

    class ThresholdAction(Action):
        def __call__(self, parser, args, values, option_string=None):
            # print '{n} {v} {o}'.format(n=args, v=values, o=option_string)
            calc, comp, threshold = values
            if not ci_in(comp, ('above', 'below')):
                raise ValueError('invalid Comparison {s!r}'.format(s=comp))
            else:
                comp = eval(comp)
            threshold = float(threshold)
            if (not hasattr(args, self.dest) or
                    getattr(args, self.dest) is None):
                setattr(args, self.dest, list())
            d = getattr(args, self.dest)
            d.append(Threshold(calc, comp, threshold))

    met_names = Suspicious.getNames(True)

    parser = ArgumentParser(prog='merge', description='Merge .results files '
                            'produced by the run_all script')
    parser.add_argument('-R', '--relative', action='store_true', help='Compute '
                        'relative values instead of absolute values')
    parser.add_argument('-r', '--recurse', nargs='?', type=int, default=False,
                        action=RecurseAction, metavar='X', help='Activates '
                        'the scripts recursive mode. This makes the script '
                        'recursively look in sub-directories of the current '
                        'directory for results files. An optional maximum '
                        'recurse limit X can be given.')
    parser.add_argument('-i', '--incl', nargs='+', metavar='DIR_ARG',
                        action='extend', default=[], type=parse_dir_arg,
                        help='Specifies particular directories to include for '
                        'the recursive option. You may optionally give a '
                        'depth with each directory in the format '
                        'DIR_ARG=\"[<depth>:]<dir name>\", where the depth is an '
                        'integer starting with 1 (the current directory). '
                        'By default a depth of \"*\" is used, indicating any '
                        'depth is valid. NOTE: the colon character (\":\") '
                        'should not appear in the directory name, but if it '
                        'must, then the depth must also be given. This option '
                        'may be specified multiple times')
    parser.add_argument('-e', '--excl', nargs='+', metavar='DIR_ARG',
                        action='extend', default=[], type=parse_dir_arg,
                        help='Specifies particular directories to exclude for '
                        'the recursive option. You may optionally give a '
                        'depth with each directory (see --incl for format).'
                        'This option may be specified multiple times')
    parser.add_argument('-t', '--tex', nargs='?', const='results.tex', metavar='FILE',
                        type=FileType('w'), help='Specifies that an additional '
                        'output file should be generated that contains the '
                        'results in a LaTeX table (in .tex format). By default '
                        'this is stored in results.tex, but an optional '
                        'filename FILE may be given')
    parser.add_argument('-p', '--perc@n', nargs='?', const='perc_at_n_results',
                        type=FileType('w'), metavar='FILE', help='Specifies '
                        'that an additional output file should be generated '
                        'that contains the percentage-at-n results. By default '
                        'this is stored in perc_at_n_results, but an optional '
                        'filename FILE may be given', dest='perc_at_n')
    parser.add_argument('-1', '--inline-perc@n', action='store_const',
                        const=None, dest='perc_at_n', help='Instead of '
                        'producing a separate percentage-at-n file, place the '
                        'results inline in the results file')
    parser.add_argument('-o', '--output', action='store', type=FileType('w'),
                        metavar='FILE', help='Store the results in the file '
                        'with filename FILE. By default the name "results" is '
                        'used', default='results')
    parser.add_argument('-d', '--decimals', action='store', type=int, default=24,
                        help='Sets the precision (number of decimal points) '
                        'for the output of all of the calculations Does not '
                        'impact percentage-at-n values. (default %(default)s, '
                        'i.e. all python-stored significance).')
    parser.add_argument('-g', '--grouping-order', choices=['metric', 'type'],
                        default='metric', dest='group',
                        help='Specifies the way in which the output should be '
                        'grouped. "metric" groups first by metrics and then by '
                        'types, "type" does the opposite (default %(default)s)')
    parser.add_argument('-c', '--calcs', nargs='+', metavar='CALC',
                        help='Specify the list of calculations to include when '
                        'merging. By default all available calculations are '
                        'included. NOTE: the names of the calculations need to '
                        'be found in the corresponding .results files')
    parser.add_argument('--threshold', nargs=3, action=ThresholdAction, default=[],
                        help='Format: --threshold <calculation> {above, below} '
                        '<float>. Specifies that an additional calculation '
                        'should be added that counts the number of versions '
                        'where the given calculation is above or below the '
                        'given float threshold. The calculations are the same '
                        'as for the --calcs argument.')
    parser.add_argument('--percentage', nargs='+', metavar='CALC',
                        help='Specify calculations that must be intepreted as '
                        'a percentage value. NOTE: the names of the '
                        'calculations need to be found in the corresponding '
                        '.results files')
    parser.add_argument('-s', '--significance', nargs='?', const='type',
                        choices=['metric', 'type'], dest='sign',
                        help='Specifies that additional significance tests '
                        'should be performed to test the differences in '
                        'results. The significance tests will either be '
                        'conducted between metrics of the same type [metric] '
                        'or between types using the same metric [type] '
                        '(default %(const)s). If type is given, and the '
                        '--tex option is also used, significance indicators '
                        'will be added to the TeX output indicating advanced '
                        'types significantly greater than their baselines '
                        '(see --base-type for changing the baseline, and '
                        '--less-significance for significantly less)')
    parser.add_argument('-f', '--flitsrs', nargs='+', metavar='METRIC',
                        help='Specify the metrics for which to display FLITSR '
                        'and FLITSR* values for. By default all metric\'s '
                        'FLITSR and FLITSR* values are shown.',
                        choices=met_names)
    parser.add_argument('-m', '--metrics', nargs='+', metavar='METRIC',
                        help='Specify the metrics to merge results for. By '
                        'default all metrics that appear in filenames of found '
                        'files will be merged. Note that this option only '
                        'restricts the output, all files available are still '
                        'read, however files not existing are not read.',
                        choices=met_names)
    parser.add_argument('-l', '--less-significance', nargs='+', metavar='CALC',
                        help='Intended for use with the --significance and '
                        '--tex options. Specify the calculations whose result '
                        'is to be tested for significantly less than the '
                        'baseline instead of significantly greater, which is '
                        'the default. Affects the significance indicators '
                        'for the TeX output. NOTE: the names of the '
                        'calculations need to be found in the corresponding '
                        '.results files', dest='sign_less', default=[])
    parser.add_argument('-b', '--base-type', help='Intended for use with '
                        '--significance and --tex options. Specify the base '
                        'type that will be compared against for all other '
                        'types when adding significance test annotations to '
                        'the TeX output. Use the format "{}_<type>" if the '
                        'baseline to compare to is dependant on the type.')
    parser.add_argument('-k', '--keep-order', action='store_true',
                        help='Instead of sorting the metrics and calculations '
                        'by alphabetical order, keep the order that the are '
                        'specified on the command line by the -m and -c '
                        'options. This option does nothing to the '
                        'corresponding order if either of those options are '
                        'unspecified')
    argcomplete.autocomplete(parser)
    args = parser.parse_args(argv)
    if ('max' not in args):
        args.max = None
    merge(args.recurse, args.max, args.incl, args.excl, args.relative,
          args.output, args.perc_at_n, args.tex, dec=args.decimals,
          group=args.group, incl_calcs=args.calcs, percs=args.percentage,
          incl_metrics=args.metrics, flitsrs=args.flitsrs, sign=args.sign,
          sign_less=args.sign_less, base_type=args.base_type,
          thrs=args.threshold, keep_order=args.keep_order)


if __name__ == "__main__":
    main()



import sys


def error(*args, **kwargs):
    """
    Print an error message and exit FLITSR. All arguments and keyword-arguments
    given are in the format of the Python print statement
    """
    print("ERROR:", end=' ', file=sys.stderr)
    print(*args, **kwargs, file=sys.stderr)
    quit()


def warning(*args, **kwargs):
    """
    Print a warning message and continue. All arguments and keyword-arguments
    given are in the format of the Python print statement
    """
    print("WARNING:", end=' ', file=sys.stderr)
    print(*args, **kwargs, file=sys.stderr)



import sys
import re
import os
from io import TextIOWrapper
from typing import Dict, Tuple, List
from flitsr.output import print_spectrum
from flitsr.split_faults import split
from flitsr.spectrum import Spectrum, Outcome
from flitsr.spectrumBuilder import SpectrumBuilder
from flitsr import errors


def construct_details(f: TextIOWrapper, method_level: bool,
                      sb: SpectrumBuilder) -> Dict[int, Spectrum.Element]:
    """
    Fills the spectrum object with elements read in from the open file 'f'.
    """
    i = 0  # number of actual lines
    method_map: Dict[int, Spectrum.Element] = {}
    methods: Dict[Tuple[str, str], Spectrum.Element] = {}
    bugs = 0
    assert f.readline() == 'name\n'  # remove header
    for line in f:
        m = re.fullmatch('([^$]+)\\$([^#]+)#([^:]+):([0-9]+)(?::(.+))?',
                         line.rstrip())
        if (m is None):
            errors.error("Incorrectly formatted line \"" + line +
                         "\" when reading input file")
        else:
            faults = []
            if (m.group(5) is not None):
                bs = m.group(5).split(':')
                if (not bs[0].isdigit()):
                    faults = [bugs]
                else:
                    faults = []
                    for b in bs:
                        faults.append(int(b))
                bugs += 1
            if (method_level):
                details = [m.group(1)+"."+m.group(2), m.group(3), m.group(4)]
                if ((details[0], details[1]) not in methods):
                    # add with first line number
                    elem = sb.addElement(details, faults)
                    methods[(details[0], details[1])] = elem
                    method_map[i] = elem
                else:
                    elem = method_map[i] = methods[(details[0], details[1])]
                    # update method's faults
                    for fault in faults:
                        if (fault not in elem.faults):
                            elem.faults.append(fault)
            else:
                details = [m.group(1)+"."+m.group(2), m.group(3), m.group(4)]
                elem = sb.addElement(details, faults)
                method_map[i] = elem
            i += 1
    return method_map


def construct_tests(tests_reader: TextIOWrapper, sb: SpectrumBuilder):
    tests_reader.readline()
    i = 0
    for line in tests_reader:
        m = re.fullmatch('([^,]+),(PASS|FAIL|ERROR)(,.*)?', line.rstrip())
        if (m is None):
            errors.error("incorrectly formatted test line in input file:",
                         line, "terminating...")
        else:
            sb.addTest(m.group(1), i, Outcome[m.group(2)])
        i += 1


def fill_spectrum(bin_file: TextIOWrapper,
                  method_map: Dict[int, Spectrum.Element],
                  sb: SpectrumBuilder):
    for t, test in enumerate(sb.get_tests()):
        line = bin_file.readline()
        if (line == ''):
            errors.error('Incorrect number of matrix lines', f'({t})',
                         'in input file, terminating...')
        arr = line.rstrip().split()
        seen = set()
        # Loop over all elements in test exec (remove test outcome at the end)
        for i in range(0, len(arr)-1):
            try:
                elem = method_map[i]
            except KeyError:
                errors.error('Incorrect number of matrix columns', f'({i})',
                             'in input file, terminating...')
            if (arr[i] != "0" and elem not in seen):
                sb.addExecution(test, elem, arr[i] != "0")
                seen.add(elem)
        # Use row to merge equivalences
        sb.split_groups_on_test(test)
    # ??? groups.sort(key=lambda group: group[0])
    # Remove groupings from spectrum
    # spectrum.remove_unnecessary()


def read_spectrum(input_path: str, split_faults: bool,
                  method_level=False) -> Spectrum:
    sb = SpectrumBuilder()
    # Getting the details of the elements
    method_map = construct_details(open(input_path+"/spectra.csv"),
                                   method_level, sb)
    # Getting the details of the tests
    construct_tests(open(input_path+"/tests.csv"), sb)
    # Constructing the spectrum
    fill_spectrum(open(input_path+"/matrix.txt"), method_map, sb)
    spectrum = sb.get_spectrum()
    del sb
    # Split fault groups if necessary
    if (split_faults):
        faults, unexposed = split(spectrum.get_faults(), spectrum)
        for elem in unexposed:
            elem.faults.clear()
            print("WARNING: Dropped faulty UUT:", elem, "due to unexposure",
                  file=sys.stderr)
        # Get element's fault lists
        fault_lists: Dict[Spectrum.Element, List[float]] = {}
        for (f_num, f_locs) in faults.items():
            for elem in f_locs:
                fault_lists.setdefault(elem, []).append(f_num)
        # Set element's fault lists
        for (elem, f_list) in fault_lists.items():
            elem.faults = f_list
        if (len(faults) == 0):
            print("ERROR: No exposable faults in ", input_path, ", terminating...",
                    sep="", file=sys.stderr)
            quit()
    return spectrum


if __name__ == "__main__":
    d = sys.argv[1]
    spectrum = read_spectrum(d, False)
    print_spectrum(spectrum)



import copy
from typing import Any, Dict, List, Tuple
from math import comb, factorial
from flitsr.spectrum import Spectrum
from flitsr.tie import Ties


def precision(n: Any, ties: Ties, spectrum: Spectrum, perc=False,
              worst_effort=False, collapse=False) -> float:
    if (len(ties.faults) == 0):
        return 0.0
    fault_num, total = method(n, ties, spectrum, perc, worst_effort, collapse)
    return fault_num/total


def recall(n: Any, ties: Ties, spectrum: Spectrum, perc=False,
           worst_effort=False, collapse=False) -> float:
    if (len(ties.faults) == 0):
        return 0.0
    fault_num, total = method(n, ties, spectrum, perc, worst_effort, collapse)
    return fault_num/len(ties.faults)


def method(n: Any, ties: Ties, spectrum: Spectrum, perc: bool,
           worst_effort: bool, collapse: bool) -> Tuple[float, int]:
    size = 0
    if (collapse):
        size = len(spectrum.groups())
    else:
        for group in spectrum.groups():
            size += len(group.get_elements())
    if (n == "b"):
        n = -1
    elif (n == "f"):
        n = len(ties.faults)
    elif (perc):
        n = n * size
    tie_iter = iter(ties)
    total = 0
    fault_num = 0.0
    try:
        while (total < n):
            tie = next(tie_iter)
            add = 0.0
            if (total+tie.len(collapse) > n and tie.num_faults() > 0):
                p = int(n - total)
                m = tie.len(collapse)
                n_l = tie.num_fault_locs(collapse)
                fpl = tie.num_faults()/n_l
                outer_top = factorial(m-p) * factorial(p)
                outer_bot = factorial(m)
                for x in range(1, p+1):
                    add += fpl*x*((comb(n_l, x) * comb(m - n_l, p - x) *
                                   outer_top)/outer_bot)
                # for i in range(curr_faults):
                #     expected_value = (i+1)*(len(uuts)+1)/(curr_faults+1)
                #     if (expected_value <= x):
                #         add += 1
                total += p
            else:
                add = tie.num_faults()
                total += tie.len(collapse)
            fault_num += add
    except StopIteration:
        pass
    return fault_num, total



class ConstIter:
    """
    An iterator that will return a constant value for a number of
    repetitions.
    """
    def __init__(self, rep, const=0):
        self.rep = rep
        self.const = const
        self.i = 0

    def __iter__(self):
        self.i = 0
        return self

    def __next__(self):
        if (self.i < self.rep):
            self.i += 1
            return self.const
        else:
            raise StopIteration



from __future__ import annotations
import pkgutil
from flitsr import advanced
from enum import Enum
import importlib
from typing import Union

_rankers = {}
_clusters = {}
_refiners = {}
all_types = {}


def register_ranker(cls):
    _rankers[cls.__name__.upper()] = cls
    all_types[cls.__name__.upper()] = cls


def register_cluster(cls):
    _clusters[cls.__name__.upper()] = cls
    all_types[cls.__name__.upper()] = cls


def register_refiner(cls):
    _refiners[cls.__name__.upper()] = cls
    all_types[cls.__name__.upper()] = cls


__all__ = [m[1] for m in pkgutil.iter_modules(advanced.__path__)]
for module in __all__:
    importlib.import_module('.'+module, package=__name__)

RefinerType = Enum('RefinerType', _refiners,  # type:ignore
                   module=advanced, qualname='advanced.RefinerType')
ClusterType = Enum('ClusterType', _clusters,  # type:ignore
                   module=advanced, qualname='advanced.ClusterType')
RankerType = Enum('RankerType', _rankers,  # type:ignore
                  module=advanced, qualname='advanced.RankerType')


class Config:
    def __init__(self, ranker: RankerType = None,
                 cluster: ClusterType = None):
        self.cluster = cluster
        self.ranker = ranker

    def __str__(self):
        return self._get_str()

    def _get_str(self, printed=False):
        if (printed):
            c = '_'
        else:
            c = '+'
        if (self.cluster is None):
            if (self.ranker is None):
                return self._get_name(RankerType['SBFL'], printed)
            else:
                return self._get_name(self.ranker, printed)
        else:
            if (self.ranker is None):
                return self._get_name(self.cluster, printed)
            else:
                return (self._get_name(self.cluster, printed) + c +
                        self._get_name(self.ranker, printed))

    def _get_name(self, typ: Union[RefinerType, ClusterType, RankerType],
                  printed) -> str:
        if (printed and hasattr(typ.value, '_print_name')):
            return typ.value._print_name.lower()
        elif (printed):
            return typ.name.lower()
        else:
            return typ.name.upper()

    def __repr__(self):
        return self._get_str()

    def get_file_name(self):
        return self._get_str(printed=True)



from __future__ import annotations
from abc import ABC, abstractmethod
from flitsr.spectrum import Spectrum
from flitsr.ranking import Ranking
from flitsr import advanced


class Ranker(ABC):
    @abstractmethod
    def rank(self, spectrum: Spectrum, base_metric: str) -> Ranking:
        pass

    def __init_subclass__(cls):
        advanced.register_ranker(cls)



from typing import Union
from os import DirEntry

class File:
    def __init__(self, file: Union[str, DirEntry]):
        fopen = open(file)
        self.lines = fopen.readlines()
        fopen.close()
        self.i = 0

    def readline(self) -> str:
        if (self.i < len(self.lines)):
            self.i += 1
            return self.lines[self.i-1]
        return ''

    def hasline(self) -> bool:
        return self.i < len(self.lines)



import numpy as np
from flitsr.suspicious import Suspicious

def explorer(activityMatrix, errorVector, numcomp, sbflMetric, numUniverse, maxUniverse, pBernoulli = 0):

    rankingList = []
    universe = []
    maxRefinement = maxUniverse
    numRef = 0
    if np.isclose(errorVector.sum(), 0):
        return rankingList
    else:
        universe.insert(0, errorVector.copy())  #insert as enqueue
        toContinue = len(universe) > 0
        while toContinue:
            e = np.array(universe.pop())    #remove as dequeue
            ranking = np.zeros((activityMatrix.shape[1], 3))
            for i in range(activityMatrix.shape[1]):
                ranking[i, 0] = i
                # get counts
                c = activityMatrix[:, i]
                tf = (e > 0).sum()
                tp = (e == 0).sum()
                ef = (e[(c > 0)] > 0).sum()
                ep = (e[(c > 0)] == 0).sum()
                ranking[i, 1] = Suspicious(ef, tf, ep, tp).execute(sbflMetric)
            u, idx = np.unique(-ranking[:, 1], return_inverse=True)
            ranking[:, 2] = idx
            ranking = ranking[(ranking[:, -1]).argsort()]
            rankingList.append(ranking)
            for i in range(min(numUniverse, numcomp)):
                if ranking[i, 1] > 0:
                    toSuppress = int(ranking[i, 0])
                    e_refined = e.copy()
                    testsToRerun = (activityMatrix[:, toSuppress] == 1)
                    for j in range(len(testsToRerun)):
                        if testsToRerun[j]:
                            e_refined[j] = np.random.binomial(1, pBernoulli, size = 1)[0]
                    if not (any([(e_refined == a_s).all() for a_s in universe])) and not (np.sum(e_refined) == 0) and numRef <= maxRefinement:
                        universe.insert(0, e_refined)
                        numRef += 1
            toContinue = len(universe) > 0 and numUniverse > 0
    return rankingList

def merge(rankingList, numcomp):

    bestScoreList = np.zeros(numcomp)
    sortedRankingList = np.zeros((numcomp, 3))

    for i in range(len(rankingList)):
        currlist = rankingList[i]
        for item in currlist:
            comp = int(item[0])
            compScore = item[1]
            if compScore > bestScoreList[comp]:
                bestScoreList[comp] = compScore


    for i in range(numcomp):
        sortedRankingList[i, 0] = i
        sortedRankingList[i, 1] = bestScoreList[i]
    u, idx = np.unique(-sortedRankingList[:, 1], return_inverse=True)
    sortedRankingList[:, 2] = idx
    sortedRankingList = sortedRankingList[(sortedRankingList[:, -1]).argsort()]

    return sortedRankingList



import sys
import os
from io import TextIOWrapper
import re
from typing import Dict, Tuple, List
from flitsr.output import print_spectrum
from flitsr.split_faults import split
from flitsr.spectrum import Spectrum, Outcome
from flitsr.spectrumBuilder import SpectrumBuilder
from flitsr import errors


def construct_details(f: TextIOWrapper, method_level: bool,
                      sb: SpectrumBuilder) -> Dict[int, Spectrum.Element]:
    """
    Fills the spectrum object with elements read in from the open file 'f'.
    """
    i = 0  # number of actual lines
    method_map: Dict[int, Spectrum.Element] = {}
    methods: Dict[Tuple[str, str], Spectrum.Element] = {}
    bugs = 0
    line = f.readline()
    while (not line == '\n'):
        m = re.fullmatch('((?:[^ ]| [^|])+)(?: \\| (.+))?', line.rstrip())
        if (m is None):
            errors.error("incorrectly formatted element line in input file:",
                         line, "terminating...")
        else:
            faults = []
            if (m.group(2) is not None):
                bs = m.group(2).split(' | ')
                if (not bs[0].isdigit()):
                    faults = [bugs]
                else:
                    faults = []
                    for b in bs:
                        faults.append(int(b))
                bugs += 1
            if (method_level):
                details_m = re.match('([^:]+):([^:]+):([0-9]+)', m.group(1))
                if (details_m is None):
                    errors.error("Incorrectly formatted line \"" + line +
                                 "\" for method level evaluation")
                else:
                    details = list(details_m.groups())
                if ((details[0], details[1]) not in methods):
                    # append with first line number
                    elem = sb.addElement(details, faults)
                    methods[(details[0], details[1])] = elem
                    method_map[i] = elem
                else:
                    elem = method_map[i] = methods[(details[0], details[1])]
                    # update method's faults
                    for fault in faults:
                        if (fault not in elem.faults):
                            elem.faults.append(fault)
            else:
                details = m.group(1).split(":")
                elem = sb.addElement(details, faults)
                method_map[i] = elem
            i += 1
            line = f.readline()
    return method_map


def construct_tests(f: TextIOWrapper, sb: SpectrumBuilder):
    line = f.readline()
    i = 0
    while (not line == '\n'):
        m = re.fullmatch('([^ ]+) (PASSED|FAILED|ERROR)( .*)?', line.rstrip())
        if (m is None):
            errors.error("incorrectly formatted test line in input file:",
                         line, "terminating...")
        else:
            sb.addTest(m.group(1), i, Outcome[m.group(2)])
        line = f.readline()
        i += 1


def fill_spectrum(f: TextIOWrapper, method_map: Dict[int, Spectrum.Element],
                  sb: SpectrumBuilder):
    for t, test in enumerate(sb.get_tests()):
        line = f.readline()
        if (line == ''):
            errors.error('Incorrect number of matrix lines', f'({t})',
                         'in input file, terminating...')
        arr = line.rstrip().split(' ')
        seen = set()
        for i in range(0, int(len(arr)/2)):
            try:
                elem = method_map[int(arr[i*2])]
            except KeyError:
                errors.error(f'Incorrect element reference "{arr[i*2]}"',
                             f'at column {i*4+1} on line {t} of the matrix',
                             'component in the input file, terminating...')
            if (elem not in seen):
                sb.addExecution(test, elem, True)
                seen.add(elem)
        # Use row to merge equivalences
        sb.split_groups_on_test(test)
    # ??? groups.sort(key=lambda group: group[0])
    # Remove groupings from spectrum
    # sb.remove_unnecessary()


def read_spectrum(input_path: str, split_faults: bool,
                  method_level=False) -> Spectrum:
    sb = SpectrumBuilder()
    method_map: Dict[int, Spectrum.Element]
    file = open(input_path)
    exec_checks = {'#tests': False, '#uuts': False, '#matrix': False}
    while (True):
        line = file.readline()
        if (line == '' or not line.startswith("#")):
            break
        elif (line.startswith("#metadata")):
            while (not line == '\n'):
                line = file.readline()
        elif (line.startswith("#tests")):
            if (exec_checks['#tests'] is True):
                print("WARNING: duplicate #tests component in input file",
                      file=sys.stderr)
            # Constructing the spectrum
            construct_tests(file, sb)
            exec_checks['#tests'] = True
        elif (line.startswith("#uuts")):
            if (exec_checks['#uuts'] is True):
                print("WARNING: duplicate #uuts component in input file",
                      file=sys.stderr)
            # Getting the details of the project
            method_map = construct_details(file, method_level, sb)
            exec_checks['#uuts'] = True
        elif (line.startswith("#matrix")):
            if (exec_checks['#matrix'] is True):
                print("WARNING: duplicate #matrix component in input file",
                      file=sys.stderr)
            # Filling the spectrum
            fill_spectrum(file, method_map, sb)
            exec_checks['#matrix'] = True
        else:
            print("ERROR: Incorrectly formatted input file at line:", line,
                  "terminating...", file=sys.stderr)
            quit()
    file.close()
    if (sum(exec_checks.values()) != 3):
        missing = [e[0] for e in exec_checks.items() if e[1] is False]
        errors.error(f"Input file missing components: {missing},",
                     "terminating...")
    spectrum = sb.get_spectrum()
    del sb
    # Split fault groups if necessary
    if (split_faults):
        faults, unexposed = split(spectrum.get_faults(), spectrum)
        for elem in unexposed:
            elem.faults.clear()
            errors.warning(f"Dropped faulty UUT: {elem} due to unexposure")
        # Get element's fault lists
        fault_lists: Dict[Spectrum.Element, List[float]] = {}
        for (f_num, f_locs) in faults.items():
            for elem in f_locs:
                fault_lists.setdefault(elem, []).append(f_num)
        # Set element's fault lists
        for (elem, f_list) in fault_lists.items():
            elem.faults = f_list
        if (len(faults) == 0):
            errors.error(f"No exposable faults in {input_path},",
                         "terminating...")
    return spectrum


if __name__ == "__main__":
    d = sys.argv[1]
    spectrum = read_spectrum(d, False)
    print_spectrum(spectrum)



import subprocess
import re
import copy
import os
from typing import List
import tempfile
from importlib import resources
from flitsr.spectrum import Spectrum
from flitsr.output import print_tcm
from flitsr.advanced.cluster import Cluster


class Parallel(Cluster):
    """
    Run one of the parallel debugging algorithms on the spectrum to
    produce multiple spectrums, and process all other options on
    each spectrum.
    """
    _parType_opts = ['bdm', 'msp', 'hwk', 'vwk']

    def __init__(self, parType: str = 'msp'):
        self.parType = parType

    def cluster(self, inp_file: str, spectrum: Spectrum,
                method_lvl=False) -> List[Spectrum]:
        # Check if Gzoltar or converted method level and if so, convert
        tmp_name = None
        if (method_lvl or (os.path.isdir(inp_file) and
                           os.path.isfile(os.path.join(inp_file, "matrix.txt")))):
            tmp_fd, tmp_name = tempfile.mkstemp(text=True)
            inp_file = tmp_name
            tmp = os.fdopen(tmp_fd, 'w')
            print_tcm(spectrum, tmp)
            tmp.close()
        # Run the parallel algorithm
        spectrums = self.partition_table(inp_file, spectrum)
        # remove temporary file if available
        if (tmp_name):
            os.remove(tmp_name)
        return spectrums

    def partition_table(self, d: str, spectrum: Spectrum) -> List[Spectrum]:
        jar_name = 'parallel-1.0-SNAPSHOT-jar-with-dependencies.jar'
        with resources.path('flitsr.advanced', jar_name) as jar_file:
            output = subprocess.check_output(['java', '-jar', str(jar_file), d,
                                              self.parType]).decode('utf-8')
        partitions = re.split("partition \\d+\n", output)[1:]
        spectrums: List[Spectrum] = []
        for partition in partitions:
            new_spectrum = copy.deepcopy(spectrum)
            tests = [int(i) for i in partition.strip().split("\n")]
            toRemove = set()
            # keep failing + all passing
            for test in new_spectrum.failing():
                if (test.index not in tests):
                    toRemove.add(test)
            for test in toRemove:
                new_spectrum.remove(test, hard=True)
            self.trim_groups(new_spectrum)
            spectrums.append(new_spectrum)
        return spectrums

    def trim_groups(self, spectrum):
        """Remove groups that are not in this block (i.e. ef = 0)"""
        toRemove = set()
        for group in spectrum.groups():
            if (spectrum.f[group] == 0):
                toRemove.add(group)
        for group in toRemove:
            spectrum.remove_group(group)



import copy
from fractions import Fraction
from itertools import permutations, chain
from math import factorial, ceil
from typing import List, Any, Optional, Set, Dict
from flitsr.spectrum import Spectrum
from flitsr.ranking import Ranking, Rank

class Tie:
    def __init__(self):
        self._elems: Set[Spectrum.Element] = set()
        self._group_len = 0
        self._num_faults = 0
        self._fault_locs = 0
        self._fault_groups = 0
        self._all_fault_locs: Dict[Spectrum.Element, Set[int]] = {}
        self._all_fault_groups: Dict[Spectrum.Element, Set[int]] = {}

    def len(self, collapse=False) -> int:
        """
        Return either the number of groups (if collapsed), or number of elements
        in this tie (if not collapsed).
        """
        if (collapse):
            return self._group_len
        else:
            return len(self._elems)

    def elems(self) -> Set[Spectrum.Element]:
        """Return the set of all the elements in this tie"""
        return self._elems

    def num_faults(self) -> int:
        """
        Return the number of unique faults found for the first time in this
        tie
        """
        return self._num_faults

    def num_fault_locs(self, collapse=False) -> int:
        """
        Return the number of faulty locations (either groups or elements) in
        this tie.
        """
        if (collapse):
            return self._fault_groups
        else:
            return self._fault_locs

    def fault_groups(self, collapse=False) -> Dict[Spectrum.Element, Set[int]]:
        """
        Return all fault locations (either groups or elements) in
        this tie, with the faults they contain.
        """
        if (collapse):
            return self._all_fault_groups
        else:
            return self._all_fault_locs


    def expected_value(self, q, weffort: bool, collapse=False) -> float:
        """
        Calculates the expected value of the qth fault in this tie. The
        expected value can either be in terms of wasted effort (not
        counting fault inspection) or actual position in the ranking.
        """
        fs = self.fault_groups(collapse)
        dups = sorted(chain(*fs.values()))
        m = self.len(collapse)
        l = self.num_fault_locs(collapse)
        nf = self.num_faults()
        if (weffort):
            expval = Fraction(m-l, l+1)
        else:
            expval = Fraction(m+1, l+1)
        # first fault or all 1 fault per loc
        if (q == 1 or (l == nf and all(len(ls) == 1 for ls in fs.values()))):
            return float(q*expval)
        elif (l > 10 or len(dups) == len(set(dups))):  # approx. solution or all single loc
            fpl = Fraction(nf, l)  # faults per location
            for i in range(1, l+1):
                if (ceil(fpl*i) >= q):
                    if (fpl*i >= q):  # exactly contained in location
                        return float(expval*i)
                    else:  # contained between this location and the next
                        return float(expval*(i + (fpl*i % 1)))
            raise ValueError(f"Could not find fault {q}")
        else:
            dist = {}
            for rank in permutations(fs.keys()):
                seen = set()
                tot = 0
                for elem in rank:
                    tot += 1
                    seen.update(fs[elem])
                    if (len(seen) >= q):
                        break
                if (tot not in dist):
                    dist[tot] = 0
                dist[tot] += 1
            frac = Fraction(sum(v*c for v, c in dist.items()), factorial(l))
            return float(expval*frac)


    def _add_group(self, group: List[Spectrum.Element],
                   faults: Dict[int, List[Spectrum.Element]],
                   seen_faults: Set[int]):
        self._elems.update(group)
        self._group_len += 1
        # Check if fault is in group
        faulty_group = False
        seen_fault_locs: Set[Spectrum.Element] = set()
        for (fault_num, fault_locs) in faults.items():
            for fault_loc in set(fault_locs).intersection(group):
                # Check & update fault number
                if (fault_num not in seen_faults):
                    self._num_faults += 1
                    seen_faults.add(fault_num)
                # Check & update faulty locs
                if (fault_loc not in seen_fault_locs):
                    seen_fault_locs.add(fault_loc)
                    self._fault_locs += 1
                    self._all_fault_locs.setdefault(fault_loc,
                                                    set()).add(fault_num)
                # Check & update faulty group
                if (not faulty_group):
                    self._fault_groups += 1
                    faulty_group = True
                    self._all_fault_groups.setdefault(group[0],
                                                      set()).add(fault_num)


class Ties:
    def __init__(self, spectrum: Spectrum, rankings: List[Ranking]):
        self.faults: Dict[int, List[Spectrum.Element]] = spectrum.get_faults()
        # needed to remove groups of fault locations
        faults = copy.deepcopy(self.faults)
        seen_faults: Set[int] = set()
        seen_groups: Set[Spectrum.Group] = set()
        self.ties: List[Tie] = []
        class RankingIter:
            def __init__(self, ranking: Ranking):
                self.r_iter = iter(ranking)
                self.cur: Optional[Rank] = next(self.r_iter, None)
            def is_active(self) -> bool:
                return self.cur is not None
            def consume(self) -> Rank:
                if (self.cur is None):
                    raise StopIteration("No more elements in ScoreIter")
                old_cur = self.cur
                self.cur = next(self.r_iter, None)
                return old_cur
            def cur_score(self):
                if (self.cur is not None):
                    return self.cur.score
                else:
                    raise StopIteration()
        r_iters = [RankingIter(r) for r in rankings]
        # Populate this Ties object
        def get_tie_groups(ri: RankingIter) -> Set[Spectrum.Group]:
            groups: Set[Spectrum.Group] = set()
            # sanity check
            if (not ri.is_active()):
                return groups
            # Get all UUTs with same score
            score = ri.cur_score()
            while (ri.is_active() and ri.cur_score() == score):
                r = ri.consume()
                groups.add(r.group)
            return groups
        while (any(ri.is_active() for ri in r_iters)):
            all_groups: Set[Spectrum.Group] = set()
            for ri in r_iters:
                if (ri.is_active()):
                    groups = get_tie_groups(ri)
                    all_groups.update(groups)
            tie = Tie()
            for group in all_groups.difference(seen_groups):
                tie._add_group(group.get_elements(), faults, seen_faults)
            seen_groups.update(all_groups)
            self.ties.append(tie)
        # Add groups not seen to bottom of tie
        not_seen = [g for g in spectrum.groups() if g not in seen_groups]
        tie = Tie()
        for group in not_seen:
            tie._add_group(group.get_elements(), faults, seen_faults)
        self.ties.append(tie)

    def __iter__(self):
        return iter(self.ties)



from __future__ import annotations
from typing import List, Dict, Any, Set, Sequence, Tuple
from bitarray import bitarray
import numpy as np
from enum import Enum
from abc import ABC, abstractmethod
from flitsr.const_iter import ConstIter
from flitsr.input_type import InputType
from flitsr.errors import error


class Outcome(Enum):
    PASSED = 0
    PASS = 0
    FAILED = 1
    FAIL = 1
    ERROR = 2


class Spectrum:
    """An implementation for a program spectrum."""
    class Test:
        """A test object holds information pertaining to a particular test."""
        def __init__(self, name: str, index: int, outcome: Outcome):
            self.name = name
            self.index = index
            self.outcome = outcome

        def __str__(self):
            return self.name

        def __repr__(self):
            return str(self)

        def __eq__(self, other):
            return self.name == other.name

        def __hash__(self):
            return hash(self.name)

    class Entity(ABC):
        @abstractmethod
        def isFaulty(self) -> bool:
            pass

        @abstractmethod
        def index(self) -> int:
            pass

    class Element(Entity):
        """
        An element object holds information pertaining to a single spectral
        element (line, method, class, etc...).
        """
        def __init__(self, details: List[str], index: int, faults: List[Any],
                     group: Spectrum.Group = None):
            if (len(details) < 1):
                raise ValueError("Unnamed element: ", *details)
            self._index = index
            self._group = group
            self.path = details[0]
            self.method = None
            self.line = None
            if (len(details) > 2):
                if (details[2].isdigit()):
                    self.line = int(details[2])
                    self.method = details[1]
                else:
                    self.method = details[1] + "|" + details[2]
            elif (len(details) > 1):
                if (not details[1].isdigit()):
                    self.method = details[1]
                else:
                    self.line = int(details[1])
            self.faults = faults
            self.tup = (self.path, self.method, self.line)
            # self.hash = hash(self.tup)

        def isFaulty(self) -> bool:
            return len(self.faults) > 0

        def index(self) -> int:
            return self._index

        def set_group(self, group: Spectrum.Group):
            self._group = group

        def group(self) -> Spectrum.Group:
            if (self._group is None):
                raise UnboundLocalError("Element group has not been set")
            return self._group

        def __str__(self):
            return "|".join(str(i) for i in self.tup if i) + \
                   (" (FAULT {})".format(",".join(str(x) for x in self.faults))
                    if self.faults else "")

        def output_str(self, type_: InputType, incl_faults=True) -> str:
            if (type_ is InputType.TCM):
                seps = ['.', ':', ':', ' | ']
            elif (type_ is InputType.GZOLTAR):
                seps = ['$', '#', ':', ':']
            else:
                error(f"Input type {type_} not supported for output string")
            gstring = ''
            path_part = self.path.rpartition('.')
            if (path_part[0] != '' and path_part[2] != ''):
                gstring = path_part[0] + seps[0] + path_part[2]
            elif (path_part[0] != '' or path_part[2] != ''):
                gstring = path_part[0] + path_part[2]
            if (self.method):
                gstring += (seps[1] if (gstring != '') else '') + self.method
            if (self.line):
                gstring += (seps[2] if (gstring != '') else '') + str(self.line)
            if (incl_faults and self.isFaulty()):
                gstring += seps[3] + seps[3].join(str(x) for x in self.faults)
            return gstring

        def __repr__(self):
            return str(self)

        def __eq__(self, other):
            return self._index == other._index

        def __hash__(self):
            return self._index

    class Group(Entity):
        """
        A spectral group is a collection of spectral elements which have
        identical spectra.
        """
        def __init__(self, elems: List[Spectrum.Element] = None,
                     index: int = None):
            """
            Create a group object. If elems is given, the group is initialized
            with the elements within, otherwise an empty group is initialized.
            """
            self._elems = elems if elems is not None else []
            self._is_faulty = any(e.isFaulty() for e in self._elems)
            self._index = index

        def append(self, element: Spectrum.Element):
            self._elems.append(element)

        def set_index(self, index: int):
            self._index = index

        def index(self) -> int:
            if (self._index is None):
                raise UnboundLocalError("Group index has not been set")
            return self._index

        def isFaulty(self) -> bool:
            return self._is_faulty

        def get_elements(self) -> List[Spectrum.Element]:
            return self._elems

        def is_in(self, element):
            return element in self._elems

        def __str__(self) -> str:
            return f"G{self._index} ({self._elems})"

        def __repr__(self) -> str:
            return str(self)

        def __eq__(self, other):
            return self._index == other._index

        def __hash__(self):
            return self._index

    class Execution():
        """
        The Execution object holds all of the spectral information pertaining
        to the execution of a particular test.
        """

        def __init__(self, test: Spectrum.Test,
                     groups: List[Spectrum.Group]):
            self._groups = groups
            self.exec = bitarray(len(self._groups))
            self.test = test

        def update(self, group: Spectrum.Group, executed: bool):
            """
            Update the execution information of a particular group
            """
            self.exec[group.index()] = executed

        def __len__(self):
            return len(self._groups)

        def __iter__(self):
            self._iter = iter(self._groups)
            return self

        def __next__(self):
            return self.get(next(self._iter))

        def __getitem__(self, elem: Spectrum.Entity) -> bool:
            if (isinstance(elem, Spectrum.Group)):
                return self.get(elem, False)
            elif (isinstance(elem, Spectrum.Element)):
                return self.element_get(elem, False)
            else:
                raise TypeError(f"Illegal Entity type: {type(elem).__name__}")

        def __setitem__(self, group: Spectrum.Group, val: bool):
            try:
                self.exec[group.index()] = val
            except KeyError:
                pass

        def get(self, group: Spectrum.Group, default: bool = False) -> bool:
            """
            Return the execution information for the given ambiguity group. If
            the element is not present in the execution, default is given.
            """
            try:
                return bool(self.exec[group.index()])
            except (KeyError, IndexError):
                return default

        def element_get(self, elem: Spectrum.Element,
                        default: bool = False) -> bool:
            """
            Return the execution information for the given element. If the
            element is not present in the execution, default is given.
            """
            # Get the element's group
            try:
                return bool(self.exec[elem.group().index()])
            except (KeyError, IndexError):
                return default

    def __init__(self, elements: List[Spectrum.Element],
                 groups: List[Spectrum.Group], tests: List[Spectrum.Test],
                 executions: Dict[Test, Dict[Element, bool]]):
        self.spectrum: Dict[Spectrum.Test, Spectrum.Execution] = {}
        self.p: Dict[Spectrum.Group, int] = dict()
        self.f: Dict[Spectrum.Group, int] = dict()
        # Initialize element related properties
        edict = {e: i for i, e in enumerate(elements)}
        self._groups = sorted(groups, key=lambda g: edict[g.get_elements()[0]])
        for i, group in enumerate(self._groups):
            group.set_index(i)
            self.p[group] = 0
            self.f[group] = 0
            for elem in group.get_elements():
                elem.set_group(group)
        self._elements: List[Spectrum.Element] = elements
        # Initialize test related properties
        self._removed: List[Spectrum.Test] = []
        self._tests: List[Spectrum.Test] = tests
        self._failing: List[Spectrum.Test] = []
        self.tp: int = 0
        self.tf: int = 0
        for test in tests:
            # Increment total counts & add to failing set
            if (test.outcome is Outcome.PASSED):
                self.tp += 1
            else:
                self._failing.append(test)
                self.tf += 1
            # Add test execution
            self.spectrum[test] = self.Execution(test, self._groups)
        # Initialize execution information
        for test in self._tests:
            test_exe = executions[test]
            seen: Set[Spectrum.Group] = set()
            for elem, exe in test_exe.items():
                group = elem.group()
                if (group not in seen):
                    seen.add(group)
                    self.spectrum[test].update(group, exe)
                    if (test.outcome is Outcome.PASSED):
                        self.p[group] += 1
                    else:
                        self.f[group] += 1

    def __getitem__(self, t: Test):
        return self.spectrum[t]

    def __iter__(self):
        self._iter = iter(self._tests)
        return self

    def __next__(self):
        return self.spectrum[next(self._iter)]

    def __len__(self):
        return len(self._tests)

    # def elements(self) -> List[Spectrum.Element]:
    #     return self._elements

    def groups(self) -> List[Spectrum.Group]:
        return self._groups

    def tests(self) -> List[Spectrum.Test]:
        return self._tests

    def failing(self) -> List[Spectrum.Test]:
        return self._failing

    def locs(self) -> int:
        return len(self._groups)

    def remove(self, test: Spectrum.Test, hard=False):
        self._tests.remove(test)
        if (test.outcome is Outcome.PASSED):
            self.tp -= 1
        else:
            self._failing.remove(test)
            self.tf -= 1
        for group in self.groups():
            self.remove_execution(test, group, hard=False)
        if (not hard):
            self._removed.append(test)

    def remove_execution(self, test: Spectrum.Test, ent: Spectrum.Entity,
                         hard=True):
        if (self.spectrum[test][ent]):
            if (isinstance(ent, Spectrum.Element)):
                group = ent.group()
            elif (isinstance(ent, Spectrum.Group)):
                group = ent
            if (hard):
                self.spectrum[test][group] = False
            if (test.outcome is Outcome.PASSED):
                self.p[group] -= 1
            else:
                self.f[group] -= 1

    def remove_group(self, group: Spectrum.Group):
        try:
            self._groups.remove(group)  # remove if there
        except ValueError:
            # Ignore if the element is not in the list
            pass
        self.p.pop(group, None)  # remove if there
        self.f.pop(group, None)  # remove if there
        # No need to remove execution (bitarray)
        # for test in self.tests():
        #     self.spectrum[test].remove_exec(rem)  # remove if there
        return group

    def reset(self):
        """Re-activates all the tests and recomputes counts"""
        for test in self._removed:
            self._tests.append(test)
            if (test.outcome is Outcome.PASSED):
                self.tp += 1
            else:
                self._failing.append(test)
                self.tf += 1
            for group in self.groups():
                if (self.spectrum[test][group]):
                    if (test.outcome is Outcome.PASSED):
                        self.p[group] += 1
                    else:
                        self.f[group] += 1
        self._removed.clear()

    def get_group(self, elem: Spectrum.Element) -> Spectrum.Group:
        """
        Given an element, return the group of elements with identical coverage
        that this element is apart of.
        """
        return elem.group()

    def get_faults(self) -> Dict[Any, List[Spectrum.Element]]:
        actual_faults: Dict[Any, List[Spectrum.Element]] = dict()
        for group in self.groups():
            for elem in group.get_elements():
                if (elem.faults):
                    for fault in elem.faults:
                        if (fault not in actual_faults):
                            actual_faults[fault] = []
                        actual_faults[fault].append(elem)
        return actual_faults

    def get_tests(self, entity: Spectrum.Entity, only_failing=False,
                  remove=False) -> Set[Spectrum.Test]:
        """
        Finds all the test cases executing the given element, and (optionally)
        removes them from the spectrum.
        """
        executing = set()
        tests = self.failing() if only_failing else self.tests()
        for test in tests:
            if (self[test][entity]):
                executing.add(test)
        if (remove):
            for test in executing:
                self.remove(test)
        return executing

    def get_executed_groups(self, test: Spectrum.Test) -> Set[Spectrum.Group]:
        return self._get_executed_entities(test, groups=True)

    def get_executed_elements(self, test: Spectrum.Test) -> Set[Spectrum.Element]:
        return self._get_executed_entities(test, groups=False)

    def _get_executed_entities(self, test: Spectrum.Test, groups=True):
        """
        Finds either the groups (default) or the elements (groups = False)
        executed in the given test.
        """
        executed = set()
        entities: Sequence[Spectrum.Entity]
        if (groups):
            entities = self._groups
        else:
            entities = self._elements
        for ent in entities:
            if (self[test][ent]):
                executed.add(ent)
        return executed

    def to_matrix(self) -> Tuple[np.ndarray, np.ndarray]:
        """
        Converts the current spectrum into a numpy matrix and error vector.
        """
        # If no matrix already, create one
        if (not hasattr(self, '_matrix')):
            # Use all test cases and most elements
            self._matrix_tests = np.array(list(self.spectrum.keys()))
            self._matrix_elems = np.array(list(self._groups))
            self._matrix = np.zeros((len(self._matrix_tests),
                                     len(self._matrix_elems)))
            self._errVector = np.zeros(len(self._matrix_tests))
            for (i, test) in enumerate(self._matrix_tests):
                for (j, elem) in enumerate(self._matrix_elems):
                    self._matrix[i][j] = self.spectrum[test][elem]
                self._errVector[i] = 0 if (test.outcome is Outcome.PASSED) else 1
        # Extract submatrix
        tmask = np.isin(self._matrix_tests, self._tests)
        emask = np.isin(self._matrix_elems, self._groups)
        matrix = self._matrix[np.ix_(tmask, emask)]
        errVector = self._errVector[tmask]
        return matrix, errVector

    def search_tests(self, name_part, incl_removed=False):
        """
        Searches for any test that matches the partial test name `name_part`.
        Returns a list of all the matches.
        """
        results = [t for t in self.tests() if t.name.find(name_part) != -1]
        if (incl_removed):
            results.extend([t for t in self._removed
                            if t.name.find(name_part) != -1])
        return results

    def search_elements(self, name_part, groups=False):
        """
        Searches for any element that matches the partial element name
        `name_part`. Returns a list of all the matches.
        """
        elems = self._groups if groups else self._elements
        results = [e for e in elems if str(e).find(name_part) != -1]
        return results



import sys
import os
from os import path as osp
from shutil import rmtree
from typing import Dict, List, Set
from flitsr.spectrum import Spectrum, Outcome
from flitsr.ranking import Ranking
from flitsr.input_type import InputType


def print_names(spectrum, ranking=None, file=sys.stdout):
    no_ranking = False
    if (ranking is None):  # make a tempoorary Scores object
        ranking = Ranking()
        for elem in spectrum.elements():
            ranking.append(elem, 0, 0)
        no_ranking = True
    for rank in ranking:
        if (no_ranking):
            print("Faulty grouping: ", "[", file=file)
        else:
            print("Faulty grouping:", rank.score, "[", file=file)
        group = rank.group
        for elem in group.get_elements():
            print(" ", elem, file=file)
        print("]", file=file)


# TODO: this function does not produce nice looking output
def print_spectrum(spectrum: Spectrum):
    for test in spectrum.tests():
        print(test.name, end=": ")
        row = spectrum[test]
        for elem in row:
            if (row[elem]):
                print(elem, end=" ")
        if (test.outcome is Outcome.PASSED):
            print('+')
        else:
            print('-')


def print_csv(spectrum: Spectrum, ranking: Ranking, file=sys.stdout):
    print("name;suspiciousness_value", file=file)
    for rank in ranking:
        group = rank.group
        for elem in group.get_elements():
            print(elem.gzoltar_str(), ';', rank.score, sep='', file=file)


def print_spectrum_csv(spectrum: Spectrum, file=sys.stdout):
    ts = [str(t.name)+' ('+('PASS' if t.outcome is Outcome.PASSED else
                            'FAIL')+')' for t in spectrum.tests()]
    print('Element', *ts, sep=',', file=file)
    # TODO: change _elements below to elements()
    for elem in spectrum._elements:
        # print(elem, end=',', file=file)
        tests = ['X' if spectrum[t][elem] else '' for t in spectrum.tests()]
        print(elem, *tests, sep=',', file=file)


def print_tcm(spectrum: Spectrum, file=sys.stdout):
    """ Output the spectrum in TCM format """
    print("#tests", file=file)
    for test in spectrum.tests():
        print(test.name, test.outcome.name, file=file)
    print(file=file)
    print("#uuts", file=file)
    # TODO: change _elements below to elements()
    for elem in spectrum._elements:
        print(elem.output_str(type_=InputType.TCM), file=file)
    print(file=file)
    print("#matrix", file=file)
    for test in spectrum.tests():
        first = True
        for elem_id, elem in enumerate(spectrum._elements):
            if (spectrum[test][elem]):
                print(("" if first else " ")+str(elem_id), "1", end="",
                      file=file)
                first = False
        print(file=file)


def print_gzoltar(spectrum: Spectrum, directory: str):
    """ Output the spectrum in Gzoltar format """
    # First check that the directory exists and is empty
    if (not osp.isdir(directory)):
        os.mkdir(directory)
    else:
        rmtree(directory)
        os.mkdir(directory)
    # Next print out the tests
    with open(osp.join(directory, "tests.csv"), 'w') as test_file:
        print("name,outcome,runtime,stacktrace", file=test_file)
        for test in spectrum.tests():
            print(test.name, "PASS" if test.outcome is Outcome.PASSED else
                  "FAIL", "", "", sep=",", file=test_file)
    with open(osp.join(directory, "spectra.csv"), 'w') as units_file:
        print("name", file=units_file)
        # TODO: change _elements below to elements()
        for elem in spectrum._elements:
            print(elem.output_str(type_=InputType.GZOLTAR), file=units_file)
    with open(osp.join(directory, "matrix.txt"), 'w') as matrix_file:
        for test in spectrum.tests():
            for elem in spectrum._elements:
                print(int(spectrum[test][elem]), end=" ", file=matrix_file)
            print('+' if test.outcome is Outcome.PASS else '-',
                  file=matrix_file)
